/*********************************************************************************************
 * (c) COPYRIGHT 2020 by NetModule AG, Switzerland
 * -------------------------------------------------------------------------------------------
 *
 * InfoCollector MODULE
 *
 * -------------------------------------------------------------------------------------------
 *
 * FILENAME      : InfoCollector.c
 * CREATION DATE : Oct 25, 2021
 * TARGET        : embedded Linux
 *
 *********************************************************************************************/

#include <stdlib.h>                      // for calloc, free
#include <string.h>                      // strerror, string functions, memcpy, ...
#include <errno.h>                       // for errno
#include <stdio.h>                       // for sscanf

#include <nmapp/exception/exception.h>
#include <nmapp/sys/assertion.h>
#include <nmapp/file/IFileOperation.h>
#include <nmapp/sys/system.h>

#include "InfoCollector-private.h"




/*********************************************************************************************
 *        LOCALS
 *********************************************************************************************/

static const char* reservedInfo = "reserved";
static const char* preEolInformation[] = {
                                            "Not Defined",  /* 0 */
                                            "Normal",       /* 1 */
                                            "Warning",      /* 2 */
                                            "Urgent"        /* 3 */
                                         };


//--------------------------------------------------------------------------------------------
void convertFwVersionToAsciiString(char* fwVersion)
{
    size_t versionLen = strlen(fwVersion);
    int i, n;
    for(i = 2, n = 0; i < versionLen; i+=2, n++)      // i=2 --> skipping the leading "0x"
    {
        unsigned int nbr = 0;
        int match = sscanf(&fwVersion[i], "%02x", &nbr);
        if((match == EOF) || (match != 1))
        {
            Throw(EX_CONVERSION, "Converting fwVersion into ascii string");
        }
        fwVersion[n] = (char)nbr;
    }
    fwVersion[n] = '\0';     // terminate string properly
}



//--------------------------------------------------------------------------------------------
uint8_t extractByteNumberFromHexString(const char* hexstring)
{
    unsigned int byteNbr = 0;
    int convertReturn = sscanf(hexstring, "0x%02x", &byteNbr);
    if((convertReturn == EOF) || (convertReturn != 1))
    {
        Throw(EX_CONVERSION, "Converting %s into byte number (return = %d)", hexstring, convertReturn);
    }
    return (uint8_t)byteNbr;
}
//--------------------------------------------------------------------------------------------
const char* getInfoFilePath(StorageDetectorRef detectorRef, const char* filename)
{
    static char filePath[2*MMC_INFO_BUFFER_SIZE] = "";
    static const size_t FILEPATH_BUF_SIZE = 2*MMC_INFO_BUFFER_SIZE;

    const char* basePath = StorageDetector_getMmcDeviceInfoPath(detectorRef);
    memset(&filePath[0], 0, FILEPATH_BUF_SIZE);
    snprintf(&filePath[0], (FILEPATH_BUF_SIZE-1), "%s/%s", basePath, filename);
    return (const char*)&filePath[0];
}
//--------------------------------------------------------------------------------------------
void readFileContentAndStripNewline(IFileOperationRef fileOpRef, const char* fileName, char* contentBuffer, uint32_t bufferSize)
{
    IFileOperation_open(fileOpRef, fileName, "r");
    IFileOperation_readFromFile(fileOpRef, contentBuffer, bufferSize, (uint32_t)0);
    IFileOperation_close(fileOpRef);
    size_t lastChar = strlen(contentBuffer) - 1;
    bool isLastCharNewLine = (contentBuffer[lastChar] == '\n');
    contentBuffer[lastChar] = isLastCharNewLine ? '\0' : contentBuffer[lastChar];
}
//--------------------------------------------------------------------------------------------
bool isReadDynamicValueDefined(const char* readValue)
{
    if(strcmp("-", readValue) == 0)
    {
        // value not available -> unsupported
        return false;
    }
    // value supported
    return true;
}

//--------------------------------------------------------------------------------------------
void getDynamicMmcData(InfoCollectorRef me)
{
    const char* filename = NULL;

    // estimated lifetime SLC and MLC:
    memset(&me->buffers.tmp[0], 0, MMC_INFO_BUFFER_SIZE);
    filename = getInfoFilePath(me->mmcDetectorRef, MMC_LIFETIME_FILENAME);
    readFileContentAndStripNewline(me->fileOpItf, filename, &me->buffers.tmp[0], MMC_INFO_BUFFER_SIZE);
    if(isReadDynamicValueDefined((const char*)&me->buffers.tmp[0]))
    {
        // make two strings out of the lifetime info:
        assert(strlen(me->buffers.tmp) == 9);
        int delimiterIdx = 4;
        me->buffers.tmp[delimiterIdx] = '\0';
        const char* slcString = (const char*)&me->buffers.tmp[0];
        const char* mlcString = (const char*)&me->buffers.tmp[delimiterIdx+1];
        // convert to numbers and match to percent (as the info is in 10% steps)
        me->dynamicInfo.lifeTime.percentUsedSlc = (int8_t)(extractByteNumberFromHexString(slcString) * 10);
        me->dynamicInfo.lifeTime.percentUsedMlc = (int8_t)(extractByteNumberFromHexString(mlcString) * 10);
        ILog_msg(me->logRef, LOG_LEVEL_DBG, "mmc lifetime estimation: SLC=%d%%...%d%%; MLC=%d%%...%d%%",
                                            (me->dynamicInfo.lifeTime.percentUsedSlc-10),
                                             me->dynamicInfo.lifeTime.percentUsedSlc,
                                            (me->dynamicInfo.lifeTime.percentUsedMlc-10),
                                             me->dynamicInfo.lifeTime.percentUsedMlc);
    }
    else
    {
        // eMMC does not support lifetime values
        static bool isLogged = false;
        me->dynamicInfo.lifeTime.percentUsedSlc = INVALID_LIFETIME_VALUE;
        me->dynamicInfo.lifeTime.percentUsedMlc = INVALID_LIFETIME_VALUE;
        if(!isLogged)
        {
            ILog_msg(me->logRef, LOG_LEVEL_INFO,
                    "The version (%s) of the current (e)MMC does not support lifetime (needs version >= 7)",
                    me->staticInfo.mmcRevision);
            isLogged = true;
        }
    }

    // pre EOL information:
    me->dynamicInfo.preEolString = NULL;
    memset(&me->buffers.tmp[0], 0, MMC_INFO_BUFFER_SIZE);
    filename = getInfoFilePath(me->mmcDetectorRef, MMC_PREEOLINFO_FILENAME);
    readFileContentAndStripNewline(me->fileOpItf, filename, &me->buffers.tmp[0], MMC_INFO_BUFFER_SIZE);
    if(isReadDynamicValueDefined((const char*)&me->buffers.tmp[0]))
    {
        me->dynamicInfo.preEolValue = (int8_t)extractByteNumberFromHexString((const char*)&me->buffers.tmp[0]);
    }
    else
    {
        me->dynamicInfo.preEolValue = 0;
    }
    bool isPreEolValueKnown = (me->dynamicInfo.preEolValue < sizeof(preEolInformation)/sizeof(const char*));
    me->dynamicInfo.preEolString = isPreEolValueKnown ? preEolInformation[me->dynamicInfo.preEolValue] : reservedInfo;
    ILog_msg(me->logRef, LOG_LEVEL_DBG, "mmc pre-EOL info = %s (%d)", me->dynamicInfo.preEolString,
                                                                      (int)me->dynamicInfo.preEolValue);

    // disk usage:
    me->diskUsageArrMB = System_getDiskUsageInfo("/");
    me->dynamicInfo.freeDiskSpaceMB = me->diskUsageArrMB[SYS_DISKUSAGE_FREE];
}
//--------------------------------------------------------------------------------------------
void getStaticMmcData(InfoCollectorRef me)
{
    const char* filename = NULL;

    // string information:
    filename = getInfoFilePath(me->mmcDetectorRef, MMC_CHIPNAME_FILENAME);
    readFileContentAndStripNewline(me->fileOpItf, filename, &me->buffers.chipName[0], MMC_INFO_BUFFER_SIZE);

    filename = getInfoFilePath(me->mmcDetectorRef, MMC_HWREVISION_FILENAME);
    readFileContentAndStripNewline(me->fileOpItf, filename, &me->buffers.hwRevision[0], MMC_INFO_BUFFER_SIZE);

    filename = getInfoFilePath(me->mmcDetectorRef, MMC_SERIALNBR_FILENAME);
    readFileContentAndStripNewline(me->fileOpItf, filename, &me->buffers.serialNbr[0], MMC_INFO_BUFFER_SIZE);

    filename = getInfoFilePath(me->mmcDetectorRef, MMC_REVISION_FILENAME);
    readFileContentAndStripNewline(me->fileOpItf, filename, &me->buffers.revision[0], MMC_INFO_BUFFER_SIZE);

    filename = getInfoFilePath(me->mmcDetectorRef, MMC_FWVERSION_FILENAME);
    readFileContentAndStripNewline(me->fileOpItf, filename, &me->buffers.fwVersion[0], MMC_INFO_BUFFER_SIZE);
    convertFwVersionToAsciiString(&me->buffers.fwVersion[0]);

    ILog_msg(me->logRef, LOG_LEVEL_DBG, "mmc data: chipName=%s, hwRev=%s, rev=%s, serial=%s, fwVersion=%s",
                                         me->staticInfo.mmcChipName, me->staticInfo.mmcHwRevision,
                                         me->staticInfo.mmcRevision, me->staticInfo.mmcSerialNbr,
                                         me->staticInfo.mmcFwVersion);

    // reliable write information:
    filename = getInfoFilePath(me->mmcDetectorRef, MMC_RELIABLEWRITE_PARAM_FILENAME);
    readFileContentAndStripNewline(me->fileOpItf, filename, &me->buffers.tmp[0], MMC_INFO_BUFFER_SIZE);
    me->staticInfo.isReliablyWritable = ((extractByteNumberFromHexString((const char*)&me->buffers.tmp[0]) & 0x01) != 0);

    filename = getInfoFilePath(me->mmcDetectorRef, MMC_RELIABLEWRITE_SETUP_FILENAME);
    readFileContentAndStripNewline(me->fileOpItf, filename, &me->buffers.tmp[0], MMC_INFO_BUFFER_SIZE);
    me->staticInfo.reliableWriteConfig.flags = extractByteNumberFromHexString((const char*)&me->buffers.tmp[0]);

    ILog_msg(me->logRef, LOG_LEVEL_DBG, "reliable write (enable=%d) config: 0x%X",
                                        (int)me->staticInfo.isReliablyWritable, me->staticInfo.reliableWriteConfig.flags);
}

/*********************************************************************************************
 *        GLOBALS
 *********************************************************************************************/
//--------------------------------------------------------------------------------------------
InfoCollectorRef InfoCollector_create(ILogRef logRef, StorageDetectorRef storageDetectorRef, IFileOperationRef fileOpRef)
{
    assert(storageDetectorRef != NULL);
    assert(fileOpRef != NULL);

    InfoCollectorRef me = (InfoCollectorRef)calloc(1, sizeof(InfoCollectorPrivate_t));
    if(me == NULL)
    {
        Throw(EX_CREATE, "Creating InfoCollector failed");
    }
    me->logRef         = logRef;
    me->mmcDetectorRef = storageDetectorRef;
    me->fileOpItf      = fileOpRef;

    me->staticInfo.mmcChipName   = &me->buffers.chipName[0];
    me->staticInfo.mmcHwRevision = &me->buffers.hwRevision[0];
    me->staticInfo.mmcRevision   = &me->buffers.revision[0];
    me->staticInfo.mmcFwVersion  = &me->buffers.fwVersion[0];
    me->staticInfo.mmcSerialNbr  = &me->buffers.serialNbr[0];

    me->dynamicInfo.lifeTime.percentUsedSlc = INVALID_LIFETIME_VALUE;
    me->dynamicInfo.lifeTime.percentUsedMlc = INVALID_LIFETIME_VALUE;
    return me;
}
//--------------------------------------------------------------------------------------------
void InfoCollector_destroy(InfoCollectorRef me)
{
    if(me != NULL)
    {
        free(me);
    }
}
//--------------------------------------------------------------------------------------------
StaticInfo_t* InfoCollector_getStaticInfo(InfoCollectorRef me)
{
    assert(me != NULL);
    getStaticMmcData(me);
    return &me->staticInfo;
}
//--------------------------------------------------------------------------------------------
DynamicInfo_t* InfoCollector_getDynamicInfo(InfoCollectorRef me)
{
    assert(me != NULL);
    getDynamicMmcData(me);
    return &me->dynamicInfo;
}
