/*********************************************************************************************
 * (c) COPYRIGHT 2021 by NetModule AG, Switzerland
 * -------------------------------------------------------------------------------------------
 *
 * Storage Information Main
 *
 * -------------------------------------------------------------------------------------------
 *
 * FILENAME      : main.c
 * CREATION DATE : Oct 25, 2021
 * TARGET        : embedded Linux
 *
 *********************************************************************************************/

#include <errno.h>
#include <stdlib.h>                    // for calloc, free, exit and its status
#include <getopt.h>                    // for options
#include <libgen.h>                    // for basename
#include <unistd.h>                    // for getpid
#include <signal.h>                    // for signal SIGSEGV
#include <execinfo.h>                  // for backtrace

#include <nmapp/exception/exception.h>
#include <nmapp/sys/assertion.h>
#include <nmapp/logging/SysLogger.h>
#include <nmapp/conversion/conversion.h>
#include <nmapp/file/FileOperation.h>
#include <nmapp/event/evt-loop.h>
#include <nmapp/process/daemonizer.h>
#include "storage-info.h"


/*********************************************************************************************
 *        LOCALS
 *********************************************************************************************/

// Constants:
//------------------------
#define        DEFAULT_APPNAME     "storage-info"

#define        DEFAULT_INFO_FILE   "/run/storage/storage0.config"

const char*    LOG_NAME_EVTLOOP        = "evtloop";
const char*    LOG_NAME_INFO_COLLECTOR = "info-collector";
const char*    LOG_NAME_INFO_WRITER    = "info-writer";


// Structures:
//------------------------
static const struct option APP_LONGOPTS[] =
{
    {"help",             no_argument,        NULL, 'h'},
    {"daemonize",        no_argument,        NULL, 'd'},
    {"pidfile",          required_argument,  NULL, 'p'},
    {"interval",         required_argument,  NULL, 'i'},
    {"file",             required_argument,  NULL, 'f'},
    SYSLOGGER_LONGOPTS,
    {NULL,        0,                  NULL,  0 }
};
static const char* APP_SHORTOPTS = "hdp:i:f:";

struct StorageInfoMain {
    StorageInfoPrivate_t  me;
    bool                  hasToDaemonize;
    const char*           pidFilePath;
    int                   pid;
    bool                  isDaemonActive;
    const char*           infoFile;
    EvtLoopRef            evtLoop;
    uint32_t              intervalTimeMS;
    StorageDetectorRef    storageDetector;
    FileOperationRef      dataFileReader;
    FileOperationRef      infoFileWriter;
};


// Variables:
//------------------------
static struct StorageInfoMain  myApp = {
                                           .me = {.appName = DEFAULT_APPNAME,
                                                  .sysLogger = NULL,
                                                  .sysLogRef = NULL,
                                                  .intervalTimer = NULL,
                                                  .infoCollector = NULL,
                                                  .infoWriter = NULL
                                                  },
                                           .hasToDaemonize = false,
                                           .pidFilePath = NULL,
                                           .pid = -1,
                                           .isDaemonActive = false,
                                           .infoFile = (const char*)DEFAULT_INFO_FILE,
                                           .evtLoop = NULL,
                                           .intervalTimeMS = (uint32_t)3600000,
                                           .storageDetector = NULL,
                                           .dataFileReader = NULL,
                                           .infoFileWriter = NULL,
                                       };




/*********************************************************************************************
 *        CALLBACK FUNCTIONS
 *********************************************************************************************/
//--------------------------------------------------------------------------------------------
static void onSegmentationFault(void* calleCtx)
{
    struct StorageInfoMain* appRef = (struct StorageInfoMain*)calleCtx;
    assert(appRef != NULL);

    enum { MAX_BACKTRACES = 16 };
    void* bt[MAX_BACKTRACES];
    int   size;
    ILog_msg(appRef->me.sysLogRef, LOG_LEVEL_EMCY, "Segmentation fault caught");
    ILog_msg(appRef->me.sysLogRef, LOG_LEVEL_EMCY, "----- Backtrace of SIGSEGV -----");
    size = backtrace(bt, MAX_BACKTRACES);
    ILog_msg(appRef->me.sysLogRef, LOG_LEVEL_EMCY, "[BT]: Found %d traces", size);
    char **btStrings = backtrace_symbols(bt, size);
    if(NULL != btStrings)
    {
        int i;
        for(i = 0; i < size; i++)
        {
            ILog_msg(appRef->me.sysLogRef, LOG_LEVEL_EMCY, "[BT]: %s", btStrings[i]);
        }
        ILog_msg(appRef->me.sysLogRef, LOG_LEVEL_EMCY, "------------------------------------------");
        free(btStrings);
        return;
    }
    ILog_msg(appRef->me.sysLogRef, LOG_LEVEL_EMCY, "[BT]: cannot malloc space for backtrace_symbols");
}
//--------------------------------------------------------------------------------------------
static void onHardTerminationSignal(struct StorageInfoMain* appRef, int signalNbr)
{
    assert(appRef != NULL);
    ILog_msg(appRef->me.sysLogRef, LOG_LEVEL_ERR, "Hard termination: Caught signal %d - %s",
                                                  signalNbr, strsignal(signalNbr));
    exit(EXIT_FAILURE);
}
//--------------------------------------------------------------------------------------------
static void onIllegalInstruction(void* calleCtx)
{
    struct StorageInfoMain* appRef = (struct StorageInfoMain*)calleCtx;
    assert(appRef != NULL);
    onHardTerminationSignal(appRef, SIGILL);
}
//--------------------------------------------------------------------------------------------
static void onAbort(void* calleCtx)
{
    struct StorageInfoMain* appRef = (struct StorageInfoMain*)calleCtx;
    assert(appRef != NULL);
    onHardTerminationSignal(appRef, SIGABRT);
}
//--------------------------------------------------------------------------------------------
static void onArithmeticOperationError(void* calleCtx)
{
    struct StorageInfoMain* appRef = (struct StorageInfoMain*)calleCtx;
    assert(appRef != NULL);
    onHardTerminationSignal(appRef, SIGFPE);
}
//--------------------------------------------------------------------------------------------
static void onBusError(void* calleCtx)
{
    struct StorageInfoMain* appRef = (struct StorageInfoMain*)calleCtx;
    assert(appRef != NULL);
    onHardTerminationSignal(appRef, SIGBUS);
}
//--------------------------------------------------------------------------------------------
static void onBadSystemCall(void* calleCtx)
{
    struct StorageInfoMain* appRef = (struct StorageInfoMain*)calleCtx;
    assert(appRef != NULL);
    onHardTerminationSignal(appRef, SIGSYS);
}
//--------------------------------------------------------------------------------------------
static void onHangup(void* calleCtx)
{
    struct StorageInfoMain* appRef = (struct StorageInfoMain*)calleCtx;
    assert(appRef != NULL);
    ILog_msg(appRef->me.sysLogRef, LOG_LEVEL_WARN, "Caught signal %d - %s", SIGHUP, strsignal(SIGHUP));
}


/*********************************************************************************************
 *        APPLICATION START-UP HELPER FUNCTIONS
 *********************************************************************************************/
//--------------------------------------------------------------------------------------------
static void printAppUsage(void)
{
    printf("Usage: %s [args]\n\n"
           "    -h | --help                            Show this help\n"
           "    -d | --daemonize                       Run as daemon\n"
           "    -p | --pidfile=path                    The PID file, see -d\n"
           "    -i | --interval=INTERVAL_MS            Sets interval=INTERVAL_MS to get the dynamic info\n"
           "                                             default=%ums\n"
           "    -f | --file=FILE                       Sets file=FILE where the information are written to\n"
           "                                             default=%s\n"
           "\n", myApp.me.appName, (unsigned int)myApp.intervalTimeMS, myApp.infoFile
    );
    printf("Used loggers: - %s\n"
           "              - %s\n"
           "              - %s\n\n",
           LOG_NAME_EVTLOOP, LOG_NAME_INFO_COLLECTOR, LOG_NAME_INFO_WRITER
          );
    SysLogCmdLine_printHelp();
}
//--------------------------------------------------------------------------------------------
void handleParsedArgument(int parsedShortArg, int argc, char **argv, int optIdx)
{
    switch (parsedShortArg)
    {
        case 'h':
            printAppUsage();
            exit(EXIT_SUCCESS);
        case 'd':
            myApp.hasToDaemonize = true;
            break;
        case 'p':
            myApp.pidFilePath = (const char*)optarg;
            break;
        case 'f':
            myApp.infoFile = (const char*)optarg;
            break;
        case 'i':
            myApp.intervalTimeMS = str2uint(optarg);
            break;
        case '?':
        default:
        {
            if(!SysLogCmdLine_isValidCommandLineOption(argv[optind - 1]))
            {
                printAppUsage();
                exit(EXIT_FAILURE);
            }
        }
    } // end switch
}
//--------------------------------------------------------------------------------------------
SysLoggerRef parseAppArgumentsAndGetLoggerConfig(int argc, char **argv)
{
    Exception_t  ex;
    SysLoggerRef loggerRef = NULL;
    int c;
    int optIdx = 0;

    Try
    {
        while(true)
        {
            c = getopt_long(argc, argv, APP_SHORTOPTS, APP_LONGOPTS, &optIdx);
            if (c == -1)
            {
                break;
            }
            handleParsedArgument(c, argc, argv, optIdx);
        }

        loggerRef = SysLogger_create(myApp.me.appName, LOG_LEVEL_INFO);
        SysLoggersCmdLine_t loggerCfg = SysLogCmdLine_parse(argc, argv);
        SysLogger_addCommandLineLoggers(&loggerCfg);
    }
    Catch(ex)
    {
        fprintf(stderr, "Failed parsing application arguments: %s", ex.msg);
        exit(EXIT_FAILURE);
    }
    return loggerRef;
}
//--------------------------------------------------------------------------------------------
void createAndSetupEventLoop(ILogRef evtLoopLogRef)
{
    myApp.evtLoop = EvtLoop_create(evtLoopLogRef, true);
    EvtLoop_registerSignalEvent(myApp.evtLoop, SIGSEGV, &onSegmentationFault, (void*)&myApp);
    EvtLoop_registerSignalEvent(myApp.evtLoop, SIGILL, &onIllegalInstruction, (void*)&myApp);
    EvtLoop_registerSignalEvent(myApp.evtLoop, SIGABRT, &onAbort, (void*)&myApp);
    EvtLoop_registerSignalEvent(myApp.evtLoop, SIGFPE, &onArithmeticOperationError, (void*)&myApp);
    EvtLoop_registerSignalEvent(myApp.evtLoop, SIGBUS, &onBusError, (void*)&myApp);
    EvtLoop_registerSignalEvent(myApp.evtLoop, SIGSYS, &onBadSystemCall, (void*)&myApp);
    EvtLoop_registerSignalEvent(myApp.evtLoop, SIGHUP, &onHangup, (void*)&myApp);
    ILog_msg(myApp.me.sysLogRef, LOG_LEVEL_DBG, "Event loop created and basically setup");
}
//--------------------------------------------------------------------------------------------
void buildingApp(void)
{
    myApp.me.sysLogRef = SysLogger_getLoggerItf(myApp.me.sysLogger);
    ILog_msg(myApp.me.sysLogRef, LOG_LEVEL_INFO, "%s starting with options: interval=%ums, file=%s",
                                                 myApp.me.appName, (unsigned int)myApp.intervalTimeMS, myApp.infoFile);
    // get (and create) logger used loggers
    ILogRef evtLoopLogRef       = SysLogger_getLoggerInstance(LOG_NAME_EVTLOOP, true);
    ILogRef infoCollectorLogRef = SysLogger_getLoggerInstance(LOG_NAME_INFO_COLLECTOR, true);
    ILogRef infoWriterLogRef    = SysLogger_getLoggerInstance(LOG_NAME_INFO_WRITER, true);

    myApp.dataFileReader = FileOperation_create(infoCollectorLogRef);
    myApp.storageDetector = StorageDetector_create(infoCollectorLogRef);
    myApp.me.infoCollector = InfoCollector_create(infoCollectorLogRef, myApp.storageDetector,
                                                  FileOperation_getItfRef(myApp.dataFileReader));

    myApp.infoFileWriter = FileOperation_create(infoWriterLogRef);
    myApp.me.infoWriter = InfoWriter_create(infoWriterLogRef,
                                            FileOperation_getItfRef(myApp.infoFileWriter), myApp.infoFile);

    createAndSetupEventLoop(evtLoopLogRef);
    myApp.me.intervalTimer = EvtLoop_createAndRegisterTimer(myApp.evtLoop, &StorageInfo_onIntervalTimeout,
                                                            myApp.intervalTimeMS, true, &myApp.me);
}
//--------------------------------------------------------------------------------------------
int cleanupApp(void)
{
    Exception_t ex;
    bool isCleaningSuccessful = true;
    Try
    {
        ILog_msg(myApp.me.sysLogRef, LOG_LEVEL_DBG, "Destroying %s %s...",
                                                    myApp.me.appName,
                                                    (myApp.isDaemonActive) ? "daemon" : "parent");

        StorageInfo_destroy(&myApp.me);
        EvtLoop_destroy(myApp.evtLoop);
        FileOperation_destroy(myApp.infoFileWriter);
        StorageDetector_destroy(myApp.storageDetector);
        FileOperation_destroy(myApp.dataFileReader);
        if(myApp.isDaemonActive && (myApp.pidFilePath != NULL))
        {
            Daemonizer_removePidFile(myApp.pidFilePath);
        }
        SysLogger_destroy(myApp.me.sysLogger);
    }
    Catch(ex)
    {
        fprintf(stderr, "Failed cleaning application: %s", ex.msg);
        isCleaningSuccessful = false;
    }
    return isCleaningSuccessful ? 0 : -1;
}
//--------------------------------------------------------------------------------------------
void daemonizeMyself()
{
    assert(myApp.pidFilePath != NULL);

    ILog_msg(myApp.me.sysLogRef, LOG_LEVEL_DBG, "Will daemonize using pid-file=%s", myApp.pidFilePath);
    myApp.pid = (int)getpid();
    pid_t forkedPid = Daemonizer_daemonize(myApp.pidFilePath);
    myApp.isDaemonActive = (forkedPid == 0);
    if(myApp.isDaemonActive)
    {
        myApp.pid = (int)getpid();
        EvtLoop_reinit(myApp.evtLoop);
        ILog_msg(myApp.me.sysLogRef, LOG_LEVEL_DBG, "%s daemonized, re-initialized event-loop and pid (=%d)",
                                                    myApp.me.appName, myApp.pid);
        return;
    }

    if(cleanupApp() != 0)
    {
        exit(EXIT_FAILURE);
    }
    exit(EXIT_SUCCESS);
}

/*********************************************************************************************
 *        GLOBALS
 *********************************************************************************************/
//--------------------------------------------------------------------------------------------
int main(int argc, char **argv)
{
    Exception_t ex;
    int         appRetVal = 0;
    myApp.me.appName = (const char*)basename(argv[0]);
    myApp.me.sysLogger = parseAppArgumentsAndGetLoggerConfig(argc, argv);
    assert(myApp.me.sysLogger != NULL);

    Try
    {
        buildingApp();

        // Initialize the created instances:
        StorageDetector_init(myApp.storageDetector);
        StorageInfo_init(&myApp.me);

        // Daemonizing:
        if(myApp.hasToDaemonize)
        {
            daemonizeMyself();
        }

        // Start Timer and Run:
        EvtLoop_startTimer(myApp.me.intervalTimer);
        EvtLoop_run(myApp.evtLoop);
    }
    Catch(ex)
    {
        ILogRef logRef = SysLogger_getLoggerItf(myApp.me.sysLogger);
        ILog_msg(logRef, LOG_LEVEL_CRIT, "%s", ex.msg);
        appRetVal = -1;
    }

    appRetVal += cleanupApp();
    myApp.me.sysLogger = NULL;
    return appRetVal;
}
